/** 
 * SALSA 1.1
 * June 2000
 * Mike Gasser (gasser@indiana.edu)
 */

// INCLUDE STATEMENTS

import java.io.*;
import java.util.Vector;
import java.awt.Color;
import java.awt.Graphics;

/**
* A Thing that has varying strength, can move, and eats Plants that
* it bumps into.  Critters occupy only one cell.
*/

public class Critter extends Thing {

  // CONSTANTS

  private static final int
    // Event types; needed for animation.
    EATING = 0,
    COLLISION = 1,
    EATING_ATTEMPT = 2,
    HITTING = 3,
    HITTING_ATTEMPT = 4,
    
    // Mouth openings
    CLOSED = 0,
    NORMAL = 1,
    WIDE = 2;
    
  // How many animation steps there are for each type of animated behavior
  // (see the integers for each above).
  public static final int[]
    N_ANIMATION_STEPS = {4, 2, 4, 2, 2};
        
  // Color constants used in painting the Critter and related Cells.
  public static final Color
    COLOR = Color.red,
    COLLISION_COLOR = Color.yellow,
    HITTING_COLOR = Color.magenta,
    HIGHLIGHT_COLOR = Color.cyan;    

  // VARIABLES
    
  private boolean
    // Whether this Critter should be highlighted.
    highlight = false,
    // Whether this Critter is gone.
    gone = false;

  private int
    // Index of the critter in world.critters.
    critterIndex,
    // Strength: determines whether Critter can perform any behaviors.
    // Direction in which the Critter moves.  Changes with turns.
    // Number of time steps the Critter has been around.
//    age = 0,
    
    // Determines whether animation happens and how.  No animation when
    // it is negative.
    animationType = -1;
    
  private Heading heading;
  
  private Strength strength;

  private Sensor[] sensors;
  
  private Effector[] effectors;
  
  private Memory memory;
  
  private int
    nRest,
    nTurn,
    nMove,
    nEat,
    nHit;

  // Index in Params.CRITTER_SPECIES
  private int
    species;

  private char
    speciesName;
    
  private Color
    strengthColor;
    
  // CONSTRUCTOR
  
  /** A new Critter starts with a random heading and is added to the
   * critters Vector in World.  It does not initially have a Cell (this
   * happens in World).
   */
  Critter (World world, UI ui, Color color, int species) {
    super (world, ui, color);
    strengthColor = color;
    setSpecies(species);
    heading = new Heading(ui);
    strength = new Strength(ui, getSpecies().getAge());
    typeString = "Critter";
    critterIndex = world.addCritter(this);
    name = typeString + critterIndex;
    edible = Params.getWorld(Params.CRITTER_EDIBILITY) > 0;
    setSensors();
    setEffectors();
    setMemory();
//    ui.write(name + " has " + memory.getNWeights() + " weights.");
  } // constructor Critter

  /**
  * Like the other constructor except that it doesn't specify the color.
  */
  Critter (World world, UI ui, int species) {
    this (world, ui, COLOR, species);
  } // constructor Critter

  // METHODS
  
  // GETTERS
  
  /**
  * Returns the Critter's value for the parameter with the given
  * index (a constant in Params).   
  */  
  private int getParam(int param) {
      return getSpecies().get(param);
  } // getParam(int)
  
  /**
  * Returns the Critter's Strength object.
  */
  public Strength getStrength() {
      return strength;
  } // getStrength()
  
  /**
  * Just returns the Critter's strength.
  */
  public int getStrengthValue() {
      return getStrength().getValue();
  } // getStrengthValue()
  
  public int getNReincarnations() {
      return getStrength().getNReincarnations();
  } // getNReincarnations()
  
  /**
  * Just returns the Critter's heading.  This should actually not be
  * accessible to the Critter itself.
  */
  public int getHeading() {
      return heading.getValue();
  } // getHeading
  
  public Memory getMemory() {
      return memory;
  } // getMemory()
  
  public char getSpeciesName() {
      return speciesName;
  }
  
  /**
  * Just returns the critter index (different from the index).
  */
  public int getCritterIndex() {
      return critterIndex;
  } // getCritterIndex
  
  /**
  * Just returns the single Cell associated with the Critter.
  */
  public Cell getCell() {
      return (getCell(0));
  } // getCell

  /**
  * Returns the Cell which would be encountered next if
  * the Critter moved one Cell in the direction of its
  * current heading.  We can probably get rid of this (it's used
  * in attack).   
  */
  public Cell getNextCell() {
      return (getCell().getNeighbor(getHeading()));
  } // getNextCell
  
  private Species getSpecies() {
      return Params.getSpecies(species);
  } // getSpecies()

  /**
  * Returns the Sensor the Critter uses.
  */
  public Sensor getSensor() {
      return sensors[0];
  } // getSensor
  
  public Sensor[] getSensors() {
      return sensors;
  } // getSensors()
  
  private Sensor getSensor(int i) {
      return sensors[i];
  } // getSensor(int)
  
  public Effector getEffector() {
      return effectors[0];
  } // getEffector()
  
  public Effector[] getEffectors() {
      return effectors;
  } // getEffectors()
  
  private int getSensorLength() {
      int l = 0;
      Sensor s;
      for (int i = 0; i < sensors.length; i++) {
          s = getSensor(i);
          if (s != null)
  	    l += s.getLength();
      }
      return l;
  } // getSensorLength()
  
  public int getAge() {
      return strength.getAge();
  } // getAge()
  
  public int getRealAge() {
      return strength.getRealAge();
  } // getRealAge()
  
  public boolean isGone() {
      return gone;
  } // isGone()

  // SETTERS
  
  private void setMemory() {
      Memory mem = getSpecies().getMemory();
//      ui.write("Setting memory for critter");
      if (mem != null) {
//	  ui.write("Species has memory.");
	  memory = mem.clone(this);
      }
      else
          memory = new Memory(ui, effectors, getSensorLength(), 0, species);
  } // setMemory
  
  private void setSensors() {
      int s;
      
      sensors = new Sensor[Params.NSENSORS];
      
      for (s = 0; s < Params.NSENSORS; s++)
	  sensors[s] = null;

      /*
      ui.write("has touch: " + getSpecies().hasSensor(Params.TOUCH) +
	       ", has vision: " + getSpecies().hasSensor(Params.VISION) +
	       ", has sonar: " + getSpecies().hasSensor(Params.SONAR));
      */
	  
      if (getSpecies().hasSensor(Params.TOUCH))
	  sensors[Params.TOUCH] = new Touch(ui, species);
     
      if (getSpecies().hasSensor(Params.VISION))
	  sensors[Params.VISION] = new Vision(ui, species);

      if (getSpecies().hasSensor(Params.SONAR))
	  sensors[Params.SONAR] = new Sonar(ui, species);

  } // setSensors()
  
  private void setEffectors() {
      nRest = getParam(Params.NREST);
      nTurn = getParam(Params.NTURN);
      nMove = getParam(Params.NMOVE);
      nEat = getParam(Params.NEAT);
      nHit = getParam(Params.NHIT);
      
      effectors = new Effector[Params.NEFFECTORS];
      effectors[Params.MOVE] = new Mover(ui, this, nMove);
      effectors[Params.TURN] = new Turner(ui, this, nTurn);
      effectors[Params.EAT] = new Eater(ui, this);

      if (nRest > 0)
          effectors[Params.REST] = new Rester(ui, this);
      else
	  effectors[Params.REST] = null;
	
      if (nHit > 0)
	  effectors[Params.HIT] = new Hitter(ui, this);
      else
          effectors[Params.HIT] = null;
  } // setEffectors()
  
  /**
  * Sets the Critter's Cell to be newCell.
  */
  public void setCell(Cell newCell) {
//      ui.write("Setting critter cell");
//      newCell.describe();
    if (cells.size() > 0)
      cells.setElementAt(newCell, 0);
    else
      cells.addElement(newCell);

    newCell.setContents(this);
  } // setCell(Cell)
  
  public void changeStrength(int change, int type) {
//    ui.write(name + " reinforced: " + change);
    strength.add(change, type);
    strengthColor = Strength.toColor(getStrengthValue(), initRGB);
  } // changeStrength(int, int)
  
  public void makeGone() {
      gone = true;
  } // makeGone()
  
  public void setSpecies(int spec) {
      species = spec;
      speciesName = (char)(getSpecies().get(Params.NAME));
      getSpecies().addMember(this);
  } // setSpecies(int)
  
  private void incrAge() {
      strength.incrAge();
  } // incrAge()

  public void writeParam(PrintWriter out, int i) {
      out.println(Params.getSpeciesKey(i) + " " + getParam(i));
  } // writeParam(PrintWriter, int)
  
  public void writeAge(PrintWriter out) {
      out.println(Params.AGE_KEY + " " + getAge());
  } // writeAge(PrintWriter)
  
  public void writeMemory(PrintWriter out) {
      getMemory().write(out);
  } // writeMemory(PrintWriter)

  /**
  * Writes the current state of the Critter to the UI.
  */
  public void describe() {
      ui.write (name + ":");
      strength.describe();
//      getCell().describe();
  } // describe
  
  // ACTING
    
  /**
  * Turns the Critter through rotation (an integer between 0 and 5)
  * and changes the strength according to the angle.
  */
  public void turn(int rotation) {
      heading.add(rotation);
      changeStrength(Params.getWorld(Params.TURN_DECR) * Heading.abs(rotation),
		     Strength.TURN);
  } // turn(int)
  
  /**
  * Moves the Critter cells forward if this is possible.  First
  * checks to see whether the Cell is empty and if it exists.
  * If it is, jumps the Critter over the intervening Cells and
  * lands it there, clearing the old Cell, making the Critter the
  * contents of the new Cell, setting the new Cell to be
  * the Critter's Cell, and decrementing the Critter's energy
  * by the appropriate amount.  Otherwise collides.
  */
  public void move(int cells) {
      Cell cell = getCell().getNthCell (cells, getHeading());
      
      if (!cell.isEmpty()) {
	  // Colliding with other object
	  ui.write (name + " failed to move " + cells +
		    " cells forward.", 1);
	  collide();
      } else {
	  reallyMove(cell, cells);
      }
  } // move(int)

  /**
  * Critter tries to eat the Thing in the next Cell, decrementing
  * strength in all cases.  If an edible Thing is there, strength is then
  * incremented by the appropriate amount.
  * The Critter doesn't go to the Cell where the eaten Thing is.   
  */
  public void eat() {
      Cell nextCell = getNextCell();
      Thing nextContents = nextCell.getContents();
      
      changeStrength(Params.getWorld(Params.EATING_DECR), Strength.EAT);
      
      if (nextContents != null && nextContents.isEdible())
	 reallyEat(nextContents, nextCell);
      else {
         setAnimation(EATING_ATTEMPT);
      }
      
      if (nextContents == null)
         ui.write(name + " wanted to eat but found nothing.", 1);
      else
	 ui.write(name + " tried to eat " + nextContents.getName(), 1);
  } // eat()

  public void hit() {
      Cell nextCell = getNextCell();
      Thing nextContents = nextCell.getContents();
      
      changeStrength(Params.getWorld(Params.HITTING_DECR), Strength.HIT);

      if (nextContents != null)
	 reallyHit(nextContents, nextCell);
      else
         setAnimation(HITTING_ATTEMPT);
      
      if (nextContents == null)
         ui.write(name + " wanted to hit but found nothing.", 1);
      else
	 ui.write(name + " tried to hit " + nextContents.getName(), 1);
  } // hit()

  /**
  * Does nothing at all.
  */
  public void rest() {
      ui.write(name + " resting.", 1);
  } // rest()
  
  /**
  * Moves the Critter to the new Cell, decrements its strength accordingly,
  * and does the associated cleanup.
  */
  private void reallyMove(Cell newCell, int ncells) {
    Cell oldCell = getCell();
    
    oldCell.clear();
    
    setCell(newCell);

    changeStrength(Params.getWorld(Params.MOVE_DECR) * ncells, Strength.MOVE);
  } // reallyMove(Cell, int)
  
  private void reallyEat(Thing food, Cell foodCell) {
      if (food instanceof Plant)
	  changeStrength(Params.getWorld(Params.PLANT_EATING_INCR),
			 Strength.PLANT_EATING);
      else
	  changeStrength(Params.getWorld(Params.CRITTER_EATING_INCR),
			 Strength.CRITTER_EATING);

      if (food instanceof Critter) {
	  ((Critter)food).changeStrength(Params.getWorld(Params.EATEN_DECR),
					 Strength.EATEN);
      } else {
        world.removeThing(food);
        foodCell.clear();
      }
      setAnimation(EATING);
  } // reallyEat(Thing, Cell)

  private void reallyHit(Thing thing, Cell foodCell) {
      thing.getHit();
      setAnimation(HITTING);
  } // reallyHit(Thing, Cell)

  /**
  * Critter collides with another thing.  Decrements the strength
  * accordingly.
  */
  private void collide() {
      ui.write ("  " + name + " collided with something.", 1);
      changeStrength(Params.getWorld(Params.COLLISION_DECR), Strength.COLLISION);
      setAnimation(COLLISION);
  } // collide()
  
  /**
  * Critters are unaffected by being hit.
  */
  public void getHit() {
//      changeStrength(Params.getWorld(Params.HIT_DECR));
  } // getHit()

  /**
  * Don't get rid of dead Critters.  They can be food.   
  */
  private void die() {
     if (!isGone()) {
       memory.dead = true;
       strength.dead = true;
       world.removeThing(this);
       getCell().clear();
     }
  } // die()
  
  private void reincarnate() {
      Memory mem = getSpecies().getMemory();

      ui.write("Reincarnating " + name);
      strength.reinit();
      strengthColor = color;
      if (mem != null)
	memory = mem.clone(this);
      else
        memory.reinit();
  } // reincarnate()
  
  /**
  * Things to do at the beginning of each run.
  */
  public void initRun() {
      strength.initRun();
  } // initRun
  
  private int[] sense() {
      int[] percept = new int[getSensorLength()];
      int s;
      int pos;
      Cell cell = getCell();
      int head = getHeading();
      Sensor sensor;
      
      for (s = 0, pos = 0; s < sensors.length; s++) {
	  sensor = getSensor(s);
	  if (sensor != null) {
	    Utils.insertVector(sensor.sense(cell, head), percept, pos);
	    pos += sensor.getLength();
	  }
      }
	  
      return percept;
  } // sense()
  
  private int getStepDecr() {
     return Params.getWorld(Params.STEP_DECR);
//      return getSpecies().getStepDecr();
  } // getStepDecr()
  
  /**
  * Causes the Critter to do what it should do in a single time step.
  * step is called in the World on each Critter.  Returns true if the
  * Critter is still alive.   
  */
  public boolean step() {

    if (getStrengthValue() > 0) {  // Only respond if strong enough

      // Get perceptual input
      int[] percept = sense();
      // getSensor().sense(getCell(), getHeading());

      // Learn in memory using the last percept and reinforcement
      if (ui.age())
        memory.learn(percept, strength.getReinforcement());

      // Reinitialize reinforcement
      strength.initStep();
      
      // Choose an action.      
      memory.respond(percept, getAge());

      // Weaken the Critter by stepDecr.
      changeStrength(getStepDecr(), Strength.STEP);

      // Increment the Critter's age.
      if (ui.age())
	  incrAge();
      
      strength.step();

      // Repaint the Critter because something must have changed
      // unless we're in step mode, in which case the whole world
      // gets repainted at once.
      if (world.frame.isShowing() && !ui.animationEnabled)
        paint();
    
      return true;

    } else {
	
	if (ui.die()) {
	    die();
	    return false;
	} else {
	    reincarnate();
	    return true;
	}
    }
  } // step()
  
  // GRAPHICS
  
 /**
  * Tells the World the number of animation steps required
  * and sets the type for this Critter.
  */
  private void setAnimation (int type) {
      world.setNAnimationSteps(N_ANIMATION_STEPS[type]);
      animationType = type;
  } // setAnimation
  
  /**
   * Various paint methods.  The details depend on the nature of the
   * Critters.  In the simplest case, they have a single background
   * color and a different color to show their heading.  In more
   * complicated cases, they have sections (sixths or thirds) that
   * show different Critter properties.  Most methods takes a openness
   * argument, specifying whether a triangle or a line is to be painted
   * for the heading.   
   */

  public void paint() {
      if (!isGone())
        paint(NORMAL);
  } // paint()
  
  public void paint(Graphics graphics) {
      if (!isGone())
        paint(graphics, NORMAL);
  } // paint(Graphics)
  
  public void paint(int openness) {
      if (!isGone())
        paint(getCell(), openness);
  } // paint(int)
  
  public void paint(Cell c) {
      if (!isGone())
        paint(c, NORMAL);
  } // paint(Cell)
  
  private void paint(Color col) {
      if (!isGone()) {
      Cell c = getCell();
      c.circleFill(col);
      
//    c.smallCircleFill(strengthColor);

      paintMouth (c, getHeading(), NORMAL);
      if (highlight)
	  c.outline(HIGHLIGHT_COLOR);
      }
  } // paint(Color)
  
  public void paint(Graphics graphics, int openness) {
      paint(graphics, getCell(), openness);
  } // paint(Graphics, int)
  
  public void paint(Cell c, int openness) {
      paint(c.getGraphics(), c, openness);
  } // paint(Cell, int)
  
  public void paint(Graphics graphics, Cell c, int openness) {
      c.circleFill(graphics, color);

      paintMouth(graphics, c, getHeading(), openness);
      c.smallCircleFill(graphics, strengthColor);

      if (speciesName != '*')
	  c.write(graphics, speciesName, Utils.getComplement(strengthColor));

      if (highlight)
        c.outline(graphics, HIGHLIGHT_COLOR);
  } // paint(Graphics, Cell, int)

  /**
  * Paints the Critter's "mouth", a triangular section of the hexagon,
  * or, in the case of CLOSED, a line.
  */
  public void paintMouth(Cell c, int direction, int openness) {
      if (openness == CLOSED)
	  c.drawRadius(getHeading());
      else
	  c.arcFill(direction);
  } // paintMouth
  
  public void paintMouth(Graphics graphics, Cell c, int direction, int openness) {
      if (openness == CLOSED)
	c.drawRadius (graphics, getHeading());
      else
	c.arcFill(graphics, direction);
  } // paintMouth
  
  /*
  private void paintTriangle() {
      Cell cell = getCell();
      cell.clear();
      cell.triangleFill(color);
  } */ // paintTriangle()
    
  /**
  * Makes this a critter to highlight.
  */
  public void setHighlight() {
      highlight = true;
      getCell().outline (HIGHLIGHT_COLOR);
  } // setHighlight
  
  /**
  * Makes this a critter to unhighlight.
  */
  public void unhighlight() {
      highlight = false;
      getCell().outline (WorldFrame.BG_COLOR);
      paint();
  } // unhighlight
  
  /**
  * Animates the Critter for one animation step.  Called in
  * World.step.   step tells which animation step this is (1 is
  * the last one).   
  */
  public void animate(int step) {
      if (animationType >= 0) {
	  // Repaint the Critter and stop animation
	  if (step == 1) {
	      paint(NORMAL);
	      animationType = -1;

	  } else if (animationType == EATING || animationType == EATING_ATTEMPT) {
	      if (step % 2 == 0)
		  paint(CLOSED);
	      else
		  paint(NORMAL);

	  } else if (animationType == COLLISION) {
	      if (step == 2)
		  paint(COLLISION_COLOR);

	  } else if (animationType == HITTING || animationType == HITTING_ATTEMPT) {
	      if (step == 2)
		  paint(HITTING_COLOR);
	  }
      }
  } // animate(int)
  
} // class Critter
