/** 
 * SALSA 1.1
 * June 2000
 * Mike Gasser (gasser@indiana.edu)
 */

import java.io.*;
import java.util.Vector;
import java.awt.Color;
import java.awt.Graphics;

/**
* An "object" appearing in the World somewhere which has its own set
* of Cells and its own color.  An abstract class, so Thing never gets
* instantiated.
*/

public abstract class Thing {

  private static final String
    CELL = "cell";
    
  protected UI ui;
  
  protected World world;
  
  protected WorldCanvas canvas;

  // The Color to be used in painting the Thing in the WorldFrame.
  protected Color
    // The initial Color (this may change for some things)
    initColor,
    // The current Color
    color;
  
  protected Vector
    // The Cells where the Thing is located.
    cells;

  // A name for the Thing and a String giving the type.
  
  protected String
    name,
    typeString = "Thing";

  protected int
    // Index of this thing in world.things
    index = 0;
    
  protected int[]
    // int values for initColor (useful when color changes)
    initRGB;
    
  protected boolean
    edible = false;
    
  /**
  * When a Thing is created, it has no Cells yet, but it gets
  * added to the World's list of Things.
  */
  Thing(World world, UI ui, Color col) {
    this.ui = ui;
    this.world = world;
    this.canvas = world.canvas;
    setColor(col);
    cells = new Vector();
    index = world.addThing(this);
  } // constructor Thing
  
  /**
  * Returns the ith Cell among those defining the Thing's location.
  */
  public Cell getCell(int i) {
    return ((Cell)(cells.elementAt(i)));
  } // method getCell
  
  /**
  * Returns the Color of the Thing.  Assumes all Cells are the
  * same Color.   
  */
  public Color getColor() {
      return color;
  } // method getColor
  
  /**
  * Returns the number of cells the Thing covers.
  */
  public int getNCells() {
     return (cells.size());
  } // method getNCells
  
  /**
  * Just returns the index of the Thing.
  */
  public int getIndex() {
      return index;
  } //method getIndex
  
  /**
  * Just returns the name of the Thing.
  */
  public String getName() {
      return name;
  } // method getName
  
  /**
  * Adds the Cell to the ones that define where the Thing is. 
  */
  public void addCell(Cell cell) {
    cells.addElement(cell);
    cell.setContents(this);
  } // method addCell
  
  /**
  * Initializes the Color of the Thing.
  */
  protected void setColor(Color c) {
      initColor = c;
      color = c;
      initRGB = new int[3];
      initRGB[0] = initColor.getRed();
      initRGB[1] = initColor.getGreen();
      initRGB[2] = initColor.getBlue();
  } // method setColor
  
  /**
  * Four methods which paint the Thing.
  */
  public void paint() {
    paint(canvas.getGraphics());
  } // method paint
  
  public void paint(Graphics graphics) {
    for (int c = 0; c < cells.size(); c++)
      paint(graphics, getCell(c));
  } // method paint
  
  public void paint(Cell c) {
      c.circleFill(color);
  } // method paint(Cell)
  
  public void paint(Graphics graphics, Cell c) {
      c.circleFill(graphics, color);
  } // method paint(Graphics, Cell)
  
  public void outline() {
      outline(canvas.getGraphics());
  }
  
  public void outline(Graphics graphics) {
      for (int c = 0; c < cells.size(); c++)
	  outline(graphics, getCell(c));
  }
  
  public void outline(Cell c) {
      c.circleOutline(color);
  }
  
  public void outline(Graphics graphics, Cell c) {
      c.circleOutline(graphics, color);
  }
  
  /**
  * Writes the current state of the Thing in the UI.  Defined
  * in subclasses of Thing.
  */
  public abstract void describe();
  // method describe

  /**
  * Writes the current state of the Thing to a file.  Defined
  * in subclasses of Thing.
  */
//  public abstract void write (PrintWriter out);
  // method write
  
  /**
  * Can Critters eat this?  Later this could depend on the
  * eater.   
  */
  public boolean isEdible() {
      return edible;
  } // isEdible()
  
  /**
  * What happens when the Thing gets hit by a Critter.
  */
  public abstract void getHit();
  // getHit()

} // class Thing
