/** 
 * SALSA 1.1
 * June 2000
 * Mike Gasser (gasser@indiana.edu)
 */

import java.io.*;
import java.awt.Graphics;
import java.awt.Color;

/**
* Subclass of the abstract class Sensor which implements a primitive
* vision system with a limited range and uses competitive learning
* to reduce the number of states to something manageable.
*/
public class Vision extends Sensor {

  public static int
    INIT_COLOR_DISTANCE = 3 * 255 * 255,
    SAMPLES_PER_PIXEL = 3;
    
  // Canvases in windows for displaying retina, competitive learning
  // network activations and weights.  Public so that paint methods
  // in canvases can be called here.

  public int
    // Index of the Critter; needed to get the right row in
    // visionCanvas.
    index = 0;
    
  private Color[][]
    pixelSamples;
    
  private Color[]
    pixels;
    
  private double[]
    // Weight how much the different samples for each pixel count.
    sampleWeights;
    
  private int
    nPixelSamples;
    
  private double
    angle;
    
  private int
    nPixels;
//    nColors;
    
 /**
  * This constructor sets the UI and the Critter, initializes
  * the retina and creates the frames (and canvases) in which
  * the retina, CL activations, and CL weights are displayed.
  */
  Vision (UI ui, int species) {
    super (ui, species);
    nPixels = getSpecies().get(Params.VISUAL_RESOLUTION);
//    nStates = Params.getNColors();
    nStates = Params.COLORS.length;
    nPixelSamples = nPixels * SAMPLES_PER_PIXEL;
    initPixels();
    angle = getSpecies().get(Params.VISUAL_ANGLE) * Math.PI / 180.0;
    length = nPixels * nStates;
  } // constructor Vision
  
  // ACCESS METHODS.
  
  public int getUnitsPerPosition() {
      return nStates;
  } // getUnitsPerPosition()
  
  public Color getPixelColor(int index) {
      return (pixels[index]);
  } // method getPixelColor(int)
  
  protected Color getDisplayColor(int index) {
      return getPixelColor(index);
  } // method getDisplayCellColor
  
  protected Color getColor(int i) {
      return Params.getColor(i);
  } // getColor(int)
  
  // SETTER METHODS.
  
  protected void setDisplayLength() {
      displayLength = getSpecies().get(Params.VISUAL_RESOLUTION);
//      displayLength = Params.getCrit(species, Params.VISUAL_RESOLUTION);
  } // method setDisplayLength
  
  private void initPixels() {
      int i, half = SAMPLES_PER_PIXEL / 2;
      double x, sum = 0.0;
      
      pixelSamples = new Color[nPixels][SAMPLES_PER_PIXEL];
      pixels = new Color[nPixels];
      sampleWeights = new double[SAMPLES_PER_PIXEL];
      for (i = 0; i < SAMPLES_PER_PIXEL; i++) {
	  x = Utils.gauss(Math.abs(half - i), 2.0);
	  sampleWeights[i] = x;
	  sum += x;
      }
      for (i = 0; i < SAMPLES_PER_PIXEL; i++) {
	  sampleWeights[i] /= sum;
//	  ui.write("Weight for " + i + ": " + sampleWeights[i]);
      }
  } // method initPixels
  
  private void setPixels() {
      for (int p = 0; p < nPixels; p++)
	  for (int s = 0; s < SAMPLES_PER_PIXEL; s++)
	      pixelSamples[p][s] = getEmptyColor();
  } // method setPixels()
  
  private Graphics getWorldGraphics() {
      return getWorldCanvas().getGraphics();
  } // method getWorldGraphics()
  
  private WorldCanvas getWorldCanvas() {
      return ui.world.canvas;
  } // method getWorldCanvas()
  
  private Color getEmptyColor() {
      return Cell.COLORS[Cell.EMPTY_COLOR];
  } // method getEmptyColor()
  
  private boolean pixelSampleEmpty(int pix, int samp) {
      return (pixelSamples[pix][samp] == getEmptyColor());
  } // method pixelSampleEmpty(int, int)
  
  private int[] getPixCoords(int x, int y, double rad, double ang) {
      int[] coords = new int[2];
      
      coords[0] = x + (int)(rad * Math.sin(ang));
      coords[1] = y - (int)(rad * Math.cos(ang));
      
      if (coords[0] < 0)
	  coords[0] += getWorldCanvas().getWidth();
      else if (coords[0] > getWorldCanvas().getWidth())
	  coords[0] -= getWorldCanvas().getWidth();
	  
      if (coords[1] < 0)
	  coords[1] += getWorldCanvas().getHeight();
      else if (coords[1] > getWorldCanvas().getHeight())
	  coords[1] -= getWorldCanvas().getHeight();
      
      return coords;
  } // method getPixCoords

  private void getPixArcs(Cell cell, int heading) {
      int[] coords;
      int x = cell.getCenterX(),
          y = cell.getCenterY(),
	  p,   // Pixel index
	  s,   // Sample index
	  D,   // Distance (3rds of Cell radius)
	  i,   // Pixel sample index
	  rad = 0; // Actual distance for the arc
	  
      double angle0 = Cell.heading2Angle(heading)
         - (angle / nPixelSamples) * ((nPixelSamples - 1) / 2),
//      - angle / 2.0,
	     angIncr = angle / nPixelSamples;
	     
//	     = angle / (nPixelSamples - 1);
      boolean anyEmpty = true;
	     
      setPixels();
      
      // Start at a distance of 3 because the closest arc overlaps with
      // the Critter's own Cell.
      D = 3;
      i = 0;

      while (anyEmpty && rad < Params.maxCellDistance) {
	  rad = (D * World.getCellRadius()) / 3;
	  anyEmpty = false;
	  for (i = 0, p = 0; p < nPixels; p++) {
	      for (s = 0; s < SAMPLES_PER_PIXEL; s++, i++) {
	      if (pixelSampleEmpty(p, s)) {
		// Not already obscured by something closer.
  	        coords = getPixCoords(x, y, rad, angle0 + i * angIncr);
	        if (coords != null) {
		    // The point is not outside the World
		    pixelSamples[p][s] = Pixel.getColor(coords, getWorldCanvas());
		    if (ui.showRays())
  		      Pixel.paint(coords, getWorldGraphics());
	        }
	      }
	      if (pixelSampleEmpty(p, s))
		  anyEmpty = true;
	      }
	  }
	  D++;
      }
  } // getPixArcs(Cell, int)
  
  private int[] pixSamplesToPixels() {
      int colorI, i, p;
      int[] sensed = new int[length];
      
      for (i = 0; i < length; i++)
	  sensed[i] = 0;

      for (p = 0; p < nPixels; p++) {
	  colorI = pixelSampleToColorInt(pixelSamples[p]);
//	  pixels[p] = COLORS[colorI];
	  pixels[p] = getColor(colorI);
	  sensed[p * nStates + colorI] = 1;
      }
      return sensed;
  } // pixSamplesToPixels()
  
  private int pixelSampleToColorInt(Color[] colors) {
      double r = 0.0, g = 0.0, b = 0.0, weight;
      int c, l = colors.length;

      for (c = 0; c < l; c++) {
	  weight = sampleWeights[c];
	  r += weight * colors[c].getRed();
	  g += weight * colors[c].getGreen();
	  b += weight * colors[c].getBlue();
      }
      return (getClosestColorInt((int)r, (int)g, (int)b));
  } // pixelSampleToColorInt(Color[])
     
  private int getClosestColorInt(int r, int g, int b) {
      int c,
          minI = 0,
          min = INIT_COLOR_DISTANCE,
	  rd, gd, bd,
	  d;
      Color color;
      
      for (c = 0; c < nStates; c++) {
//	  color = COLORS[c];
	  color = getColor(c);
	  rd = color.getRed() - r;
	  gd = color.getGreen() - g;
	  bd = color.getBlue() - b;
	  d = rd * rd + gd * gd + bd * bd;
	  if (d < min) {
	      min = d;
	      minI = c;
	  }
      }
      return minI;
  } // getClosestColorInt(int, int, int)
  
  public int[] sense(Cell cell, int direction) {
      int[] sensed;
      
      getPixArcs(cell, direction);
      sensed = pixSamplesToPixels();
      
      return sensed;
  } // sense(Cell, int)
  
  protected void paint(Graphics g, int i, int x, int y, int w, int h) {
      g.fillRect(x + 1, y + 1, w - 1, h - 1);
  } // paint(Graphics, int, int, int, int, int)

  protected void clear(Graphics g, int x, int y, int w, int h) {}

} // class Vision
