/** 
 * SALSA 2
 * July 2000
 * Mike Gasser (gasser@indiana.edu)
 */

import java.awt.*;
import java.awt.event.*;

public class WeightCanvas extends CritterCanvas {
    
  public final static double
    MAX_MIN_OFFSET = .05;

  protected static final int
    NAMES_WIDTH = 13,
    MAX_POSITIONS = 6,
    LEGEND_HEIGHT = 57,
    WIDTH = 800,
    HEIGHT = 260;
    
  protected int
    nEffectors,
    effectorWidth,
    cellsPerEffector,
    halfCellHeight,
    barHeight,
    legendHeight,
    legendDivision;
    
  protected int[] starts;
  
  protected Memory[] memories;
  protected Sensor[][] sensors;
  protected Effector[][] effectors;

  WeightCanvas (UI ui, World world) {
      super(ui, world);
  } // constructor WeightCanvas
  
  protected void init() {
      setMemories();
      sensors = getSensors();
      effectors = getEffectors();
  }
  
  /**
  * Has to precede setDimensions().
  */
  private void setMemories() {
      this.memories = getMemories();
      for (int c = 0; c < nCritters; c++)
	  memories[c].setWeightCanvas(this);
  } // method setMemories()
  
  /**
  * First figures the number of cells for each Critter.
  * Then figures the Canvas width based on the Critter with the most cells.
  * Then goes back and figures the cellWidths for each Critter.   
  */
  protected void setDimensions() {
      Memory memory;
      int c;
      int maxNCells = 0;
      int maxCellWidth;
      
      starts = new int[Params.NEFFECTORS];

      namesWidth = NAMES_WIDTH;

      cellHeight = HEIGHT / nCritters;
      halfCellHeight = cellHeight / 2;
      barHeight = cellHeight / 3;
      mainHeight = cellHeight * nCritters + 1;
      legendHeight = LEGEND_HEIGHT;
      legendDivision = 13;
      height = mainHeight + legendHeight + 1;
            
      nEffectors = Species.getMax(Params.NREST) +
		   Species.getMax(Params.NHIT) +
		   Species.getMax(Params.NEAT) +
		   Species.getMax(Params.NMOVE) +
		   Species.getMax(Params.NTURN);
		   
      cellsPerEffector = MAX_POSITIONS * Params.NSENSORS;
      
      cellWidth = WIDTH / (nEffectors * cellsPerEffector);
      
      effectorWidth = cellWidth * cellsPerEffector;
      
      mainWidth = effectorWidth * nEffectors;
      width = mainWidth + namesWidth + 1;
      
      starts[Params.REST] = 0;
      starts[Params.EAT] = starts[Params.REST] +
        Species.getMax(Params.NREST) * effectorWidth;
      starts[Params.MOVE] = starts[Params.EAT] +
        Species.getMax(Params.NEAT) * effectorWidth;
      starts[Params.HIT] = starts[Params.MOVE] +
        Species.getMax(Params.NMOVE) * effectorWidth;
      starts[Params.TURN] = starts[Params.HIT] +
        Species.getMax(Params.NHIT) * effectorWidth;

  } // method setDimensions()
  
  /**
  * Clear the canvas and draw the unit boundaries.
  */
  /*
  public void clear() {
    int critIndex;
    Graphics g = getGraphics();
      
    g.clearRect (0, 0, width, height);
    g.setColor (BORDER_COLOR);
    
    for (critIndex = 0; critIndex < nCritters; critIndex++) {      
      g.drawRect (0, 2 * critIndex * cellHeight, width, 2 * cellHeight);
    }
  }*/ // method clear
  
  public void paintCritter(int critIndex) {
      paintCritter(getGraphics(), critIndex);
  } // method paintCritter(int)
  
  public void paintCritter (Graphics g, int critIndex) {
      if (!isDead(critIndex)) {
	  Memory memory = memories[critIndex];
	  Sensor[] critSensors = sensors[critIndex];
	  Sensor sensor;
	  Effector[] critEffectors = effectors[critIndex];
	  Effector effector;
	  double[] weights;
	  int e, o, i, s, x, e1;
	  int y = cellHeight * critIndex;
	  
//	  int[] highLow;
	  
	  for (e = 0, o = 0; e < Params.NEFFECTORS; e++) {
	      effector = critEffectors[e];
	      x = starts[e];
	      if (effector != null) {
	          for (e1 = 0; e1 < effector.getNActions(); e1++, o++) {
		      weights = memory.getIn2Out(o);
		      // i the input unit for the beginning of sensor s.
		      for (s = 0, i = 0; s < Params.NSENSORS; s++, x += cellWidth * MAX_POSITIONS) {
			  sensor = critSensors[s];
			  if (sensor != null) {
			      paintWeights(g, x, y, weights, sensor, i);
			      i += sensor.getLength();
			  }
		      }
	          }
	      }
	  }
      }
  } // method paintCritter(Graphics, int)
  
  private void paintWeights(Graphics g, int startX, int y,
			    double[] weights, Sensor sensor, int start) {
      int distinctions = sensor.getUnitsPerPosition();
      int position;
      int i, j, x;
      int maxI = start, minI = start;
      double sum, maxWt = -100.0, minWt = 100.0;
      double wt;
      
      for (position = 0, i = start, x = startX;
	   position < sensor.getDisplayLength();
	   position++, x += cellWidth) {
	  // Find the max and min for this position
	  sum = 0.0;
	  maxWt = -100.0;
	  minWt = 100.0;
	  for (j = 0; j < distinctions; j++, i++) {
	      wt = weights[i];
//	      ui.write("Weight: " + wt);
	      sum += wt;
	      if (wt > maxWt) {
		  maxWt = wt;
		  maxI = j;
	      }
	      if (wt < minWt) {
	          minWt = wt;
	          minI = j;
	      }
	  }
          sum /= distinctions;
//	  ui.write("Sum: " + sum);
//	  ui.write("Max: " + maxWt + ", min: " + minWt);
	  if (maxWt <= 0.0 || maxWt < sum + MAX_MIN_OFFSET)
	      maxI = -1;
	  if (minWt >= 0.0 || minWt > sum - MAX_MIN_OFFSET)
	      minI = -1;
	  paintWeights(g, x, y, sensor, maxI, minI);
      }
  } // paintWeights(Graphics, int, double[], Sensor, int)
  
  private void paintWeights(Graphics g, int x, int y, Sensor sensor, int maxI, int minI) {
      if (maxI >= 0)
        g.setColor(sensor.getColor(maxI));
      else
	g.setColor(BACKGROUND_COLOR);

      g.fillRect(x + 1, y + 1, cellWidth - 1, halfCellHeight - 1);

      if (minI >= 0)
        g.setColor(sensor.getColor(minI));
      else
	g.setColor(BACKGROUND_COLOR);
	
      g.fillRect(x + 1, y + halfCellHeight + 1, cellWidth - 1, halfCellHeight - 1);
  } // paintWeights(Graphics, int, int, Sensor, int, int)
  
  protected void drawLines(Graphics g) {
    int i, x, c;
    
    g.setColor(BORDER_COLOR4);
    for (i = 0; i < nEffectors * 18; i++) {
	x = i * cellWidth;
	g.drawLine(x, 0, x, mainHeight - 1);
    }

    g.setColor(BORDER_COLOR3);
    for (i = 0; i < nEffectors * 3; i++) {
	x = i * (cellWidth * 6);
	g.drawLine(x, 0, x, mainHeight - 1);
    }
    for (c = 0; c < nCritters; c++)
	g.drawLine(1, (c * cellHeight) + halfCellHeight,
		    mainWidth, (c * cellHeight) + halfCellHeight);

    g.setColor(BORDER_COLOR);
    g.drawLine(0, 0, width - 1, 0);
    g.drawLine(0, 0, 0, height - 1);
    g.drawLine(0, height - 1, width - 1, height - 1);
    g.drawLine(width - 1, 0, width - 1, height - 1);
    g.drawLine(starts[Params.TURN], 0, starts[Params.TURN], mainHeight + legendDivision);
    g.drawLine(starts[Params.MOVE], 0, starts[Params.MOVE], mainHeight + legendDivision);
    g.drawLine(starts[Params.EAT], 0, starts[Params.EAT], mainHeight + legendDivision);
    g.drawLine(starts[Params.HIT], 0, starts[Params.HIT], mainHeight + legendDivision);
    g.drawLine(1, mainHeight + legendDivision, width - 1, mainHeight + legendDivision);
    g.drawLine(mainWidth, 0, mainWidth, height - 1);
    
    g.setColor(BORDER_COLOR2);
    for (i = 1; i < Species.getMax(Params.NTURN); i++) {
	x = starts[Params.TURN] + i * effectorWidth;
	g.drawLine(x, 0, x, mainHeight - 1);
    }
    for (i = 1; i < Species.getMax(Params.NMOVE); i++) {
	x = starts[Params.MOVE] + i * effectorWidth;
	g.drawLine(x, 0, x, mainHeight - 1);
    }
  } // method drawLines()
  
  protected void writeLegend(Graphics g) {
      int i;
      int y = mainHeight + 11;
      int x;
      
      g.setColor(Color.black);
      g.setFont(FONT);
      g.drawString(Params.EFFECTOR_LABELS[Params.TURN], starts[Params.TURN] + 2, y);
      g.drawString(Params.EFFECTOR_LABELS[Params.EAT], starts[Params.EAT] + 2, y);
      g.drawString(Params.EFFECTOR_LABELS[Params.MOVE], starts[Params.MOVE] + 2, y);
      g.drawString(Params.EFFECTOR_LABELS[Params.REST], starts[Params.REST] + 2, y);
      g.drawString(Params.EFFECTOR_LABELS[Params.HIT], starts[Params.HIT] + 2, y);
      
      g.drawString(Params.SENSOR_LABELS[Params.TOUCH], 8, height - 2);
      g.drawString(Params.SENSOR_LABELS[Params.VISION], 8 + 48, height - 2);
      g.drawString(Params.SENSOR_LABELS[Params.SONAR], 8 + 96, height - 2);
      
      y = mainHeight + legendDivision + 7;
      
      g.drawString("Maximum weight", 160, y + 11);
      g.drawString("Minimum weight", 160, y + 23);
      
    g.setColor(BORDER_COLOR4);
    for (i = 0, x = 8; i < 18; i++, x += 8)
	g.drawLine(x, y, x, y + 24);
    
    g.setColor(BORDER_COLOR3);
    for (i = 0, x = 56; i < 2; i++, x += 48) {
	g.drawLine(x, y, x, y + 24);
    }
    g.drawLine(8, y + 12, 152, y + 12);

    g.setColor(BORDER_COLOR);
    g.drawLine(8, y, 8, y + 24);
    g.drawLine(8, y, 152, y);
    g.drawLine(152, y, 152, y + 24);
    g.drawLine(8, y + 24, 152, y + 24);
    
  } // writeLegend(Graphics)
  
  /*
  private int[] getHighLow(double[] weights, int start, int distance) {
      int i;
      double highest = -100.0;
      double lowest = 100.0;
      double weight;
      int highestI = start;
      int lowestI = start;
      int[] highLow = new int[2];
      
      for (i = start; i < start + distance; i++) {
	  weight = weights[i];
	  if (weight > highest) {
	      highest = weight;
	      highestI = i;
	  }
	  if (weight < lowest) {
	      lowest = weight;
	      lowestI = i;
	  }
      }
      highLow[0] = highestI;
      highLow[1] = lowestI;
      
      return highLow;
  }*/ // getHighLow(double[], int, int)

} // class WeightCanvas