/** 
 * SALSA 2
 * July 2000
 * Mike Gasser (gasser@indiana.edu)
 */

import java.io.*;
import java.util.Vector;
import java.awt.*;

/**
* The World, consisting of a grid of hexagonal Cells,
* in which all of the Things exist.  The current state of the
* World is drawn in the WorldCanvas inside the WorldFrame, both
* of which are created when World is instantiated.  World is
* Runnable so that the Critters in it can be run for a fixed number
* of time steps with pauses between time steps.
*/

public class World implements Runnable {

  // Color constants for showing grid and collisions.
  public final static Color
    GRID_COLOR = Color.darkGray,
    HIGHLIGHT_COLOR = Color.cyan,
    COLLISION_COLOR = Color.yellow;
    
  public final static double
    // Double constant to help in calculating Cell dimensions.
    SQRT_THREE = Math.sqrt(3.0),
    // Double constant to implement noise in various places.
    NOISE_PROB = 0.0;
    
  public final static int
    STM = 0,
    LTM = 1,
    STRENGTH = 2,
    REINFORCEMENT = 3,
    SEARCH_LIMIT = 1000,
    SHOW_STATS_FREQ = 5;

  private static int
    cellIncr;

  // INSTANCE VARIABLES
    
  // UI, where messages get written to.
  private UI ui;
  
  // Where the world gets drawn.
  public WorldFrame frame;
  public WorldCanvas canvas;

  private String
    // Name of the world
    name;
    
  private String
    // File and directory for last setup file loaded.
    setupFile = Params.FILE,
    setupDirectory = UI.getDataDirectory();

  private Vector
    fertileCells = new Vector(),
    // Vectors of Things, Critters, Plants in the World.
    things = new Vector(),
    critters = new Vector(),
    plants = new Vector();
    
  // Thread allowing the World to be run in discrete steps with
  // fixed pauses between them.
  private Thread thread;
  
  private int
    // Number of time steps to run animation.  Set by Critters that need
    // to be animated.
    nAnimationSteps = 0,
    nPlants = 0,
    // Number of time steps since first run began.
    totalTimeSteps = 0;
  
  private boolean
    userPlacePlants = false,
    // Whether to describe Critters after each run.
    describeCritters = true,
    // Whether to keep the number of Plants constant.
    maintainNPlants = true,
    // Whether the World has already been filled with Things.
    isInitialized = false;

  // The Cells in the World.
  private Cell cells[][];
  
  private CritterFrame[] critterFrames = {null, null, null, null};
  
//  private MemoryFrame memoryFrame = null;
//  private WeightFrame weightFrame = null;
//  private StatsFrame statsFrame = null;
//  private ReinfFrame reinfFrame = null;
  
  /**
  * When World is instantiated, its WorldFrame and WorldCanvas are
  * created, the World is filled with cells, and the initial Things
  * are created.   
  */
  World (UI ui) {
      this.ui = ui;
      name = "World";
      cellIncr = (int)(UI.cellSize * SQRT_THREE);
      doCells();
      initFrame();
//      frame = new WorldFrame (this, ui, name,
//			      Params.getWorld(Params.NCOLUMNS), Params.getWorld(Params.NROWS),
//			      UI.cellSize, cellIncr);
////      frame.setVisible (true);
      createThings(false);
//      initFrames(false);
//      ui.write ("Created world: " + Params.getWorld(Params.CELL_SIZE) + ", " + Params.getWorld(Params.NCOLUMNS) + ", "
//      + Params.getWorld(Params.NROWS));
  } // constructor World
  
  public void initFrames(boolean remove, boolean create) {
      if (remove) {
	  removeFrames();
//	  ui.removeWindow(memoryFrame);
//	  ui.removeWindow(statsFrame);
//	  ui.removeWindow(reinfFrame);
//	  ui.removeWindow(weightFrame);
      }
      if (create) {
	  int worldHeight = frame.getSize().height / 2;
	  int xOffset = 0;

	  if (ui.showSTM())
//	  (Params.getDisplay(Params.SHOW_STM) == 1)
	  {
	      setCritterFrame(STM, new MemoryFrame(ui, this, "STM"));
	      getCritterFrame(STM).setLocation(0, worldHeight);
	      xOffset += getCritterFrame(STM).getSize().width;
	  }
	  
	  if (ui.showStrengths())
//        (Params.getDisplay(Params.SHOW_STRENGTHS) == 1)
	  {
	      setCritterFrame(STRENGTH, new StatsFrame(ui, this, "Strength & Age"));
	      getCritterFrame(STRENGTH).setLocation(xOffset, worldHeight);
	      xOffset += getCritterFrame(STRENGTH).getSize().width;
	  }
	  if (ui.showReinforcements())
//        (Params.getDisplay(Params.SHOW_REINFORCEMENT) == 1)
	  {
	      setCritterFrame(REINFORCEMENT, new ReinfFrame(ui, this, "Reinforcement"));
	      getCritterFrame(REINFORCEMENT).setLocation(xOffset, worldHeight);
	      xOffset += getCritterFrame(REINFORCEMENT).getSize().width;
	  }
  	  if (ui.showLTM())
//        (Params.getDisplay(Params.SHOW_LTM) == 1)
	  {
	      setCritterFrame(LTM, new WeightFrame(ui, this, "LTM"));
	      getCritterFrame(LTM).setLocation(0, worldHeight + worldHeight / 2);
  	  }
      }
  } // initFrames()
  
  private void removeFrames() {
      for (int i = 0; i < critterFrames.length; i++)
	  ui.removeWindow(getCritterFrame(i));
  }

  /*
  public Frame getSTMFrame() {
      return getCritterFrame(STM);
  }
  public Frame getLTMFrame() {
      return getCritterFrame(LTM);
  }
  public Frame getStrengthFrame() {
      return getCritterFrame(STRENGTH);
  }
  public Frame getReinfFrame() {
      return getCritterFrame(REINFORCEMENT);
  }
  */
  
  private void initCritterFrame(int i) {
      if (i == LTM)
	  setCritterFrame(LTM, new WeightFrame(ui, this, "LTM"));
      else if (i == STM)
	  setCritterFrame(STM, new MemoryFrame(ui, this, "STM"));
      else if (i == STRENGTH)
	  setCritterFrame(STRENGTH, new StatsFrame(ui, this, "Strength & Age"));
      else if (i == REINFORCEMENT)
	  setCritterFrame(REINFORCEMENT, new ReinfFrame(ui, this, "Reinforcement"));
  }
  
  public CritterFrame getCritterFrame(int i) {
      return critterFrames[i];
  }
  
  private void setCritterFrame(int i, CritterFrame f) {
      critterFrames[i] = f;
  }
  
  private void repaintCritterFrame(int i) {
      CritterFrame cf = getCritterFrame(i);
      if (cf != null)
	  cf.repaint();
  }
  
  public void setVisibleCritterFrame(int i, boolean show) {
      CritterFrame cf = getCritterFrame(i);
      if (show && cf == null)
	  initCritterFrame(i);
      else
	  cf.setVisible(show);
  } // setVisibleCritterFrame(int, boolean)

  public boolean isInit() {
      return isInitialized;
  } // isInit()
  
  public void setInit() {
      isInitialized = true;
  } // setInit()
  
  public void reinit() {
      reinit(setupDirectory, setupFile);
  } // reinit()
  
  public void reinit(Vector remThings, Vector remFertileCells,
		     Vector[] newCells) {
      int i, j;
      Plant p;
      Vector v;
      
      removeThings(remThings);

      removeFertileCells(remFertileCells);
      
      v = newCells[EditWorldFrame.ROCK];
      // Create Rocks in the newRockCells
      for (i = 0; i < v.size(); i++)
	  createRock((Cell)(v.elementAt(i)));
      // Create Plants in the newPlantCells

      v = newCells[EditWorldFrame.TENDER_PLANT];
      for (i = 0; i < v.size(); i++) {
	  p = createPlant((Cell)(v.elementAt(i)));
	  p.setToughness(false);
      }

      v = newCells[EditWorldFrame.TOUGH_PLANT];
      for (i = 0; i < v.size(); i++) {
	  p = createPlant((Cell)(v.elementAt(i)));
	  p.setToughness(true);
      }

      v = newCells[EditWorldFrame.FERTILE_CELL];
      addFertileCells(v);

      if (newCells.length >= EditWorldFrame.CRITTER) {
	  // One critter vector for each species
	  for (i = 0; i < Params.getNSpecies(); i++) {
	      v = newCells[EditWorldFrame.CRITTER + i];
	      for (j = 0; j < v.size(); j++) {
		  new Critter(this, ui, i).setCell((Cell)(v.elementAt(j)));
	      }
	  }
          initFrames(true, true);
      }
      // Reset the number of plants to be however many are left.
      // (Don't bother with the number of rocks).
      Params.setWorld(Params.NPLANTS, Params.VALUE, nPlants);
  } // reinit(Vector, Vector)
  
  private void removeThings(Vector remThings) {
      for (int t = 0; t < remThings.size(); t++)
	  removeThing((Thing)(remThings.elementAt(t)));
  } // removeThings(Vector)
  
  /**
  * Clear out the Vectors critters and things,
  * empties all Cells, and recreates the wall.  Gets rid
  * of windows and creates new ones.
  */
  public void reinit(String directory, String paramFile) {
    int r, c;
    int oldNRows = Params.NROWS;
    int oldNColumns = Params.NCOLUMNS;
    
    setupDirectory = directory;
    setupFile = paramFile;

    critters.removeAllElements();
    plants.removeAllElements();
    things.removeAllElements();
    reinitFertileCells();

    nPlants = 0;

    Params.reinit();

    Params.read(ui, directory, paramFile);

    if ((Params.getWorld(Params.NROWS) != oldNRows) ||
	(Params.getWorld(Params.NCOLUMNS)) != oldNColumns) {
	ui.removeWindow(frame);
	doCells();
	initFrame();
    } else {
	for (r = 0; r < Params.getWorld(Params.NROWS); r++) {
    	  for (c = 0; c < Params.getWorld(Params.NCOLUMNS); c++) {
	    getCell(c, r).setContents (null);
    	  }
	}
    }

    if (frame.isShowing())
      clear();
  
    createThings(true);

    paint();

    totalTimeSteps = 0;
    
    System.gc();
    
//    initFrames(true);

  } // reinit(String)
  
  private void initFrame() {
	frame = new WorldFrame(this, ui, name,
			      Params.getWorld(Params.NCOLUMNS),
			      Params.getWorld(Params.NROWS),
			      UI.cellSize, cellIncr);
	ui.addWindow(frame);
	frame.setLocation(ui.getSize().width, 0);
  } // initFrame()
  
  public void setFertileCells(Vector v) {
      int c;
      Cell cell;
      fertileCells = (Vector)(v.clone());
      for (c = 0; c < getNFertileCells(); c++) {
	  getFertileCell(c).setFertile(true);
      }
  } // setFertileCells(Vector)
  
  private void addFertileCells(Vector v) {
      int i;
      Cell cell;
      for (i = 0; i < v.size(); i++) {
	  cell = (Cell)(v.elementAt(i));
	  fertileCells.addElement(cell);
	  cell.setFertile(true);
      }
  } // addFertileCells(Vector)
  
  private void removeFertileCells(Vector v) {
      int i;
      Cell cell;
      for (i = 0; i < v.size(); i++) {
	  cell = (Cell)(v.elementAt(i));
	  fertileCells.removeElement(cell);
	  cell.setFertile(false);
      }
  } // removeFertileCells(Vector)

  /**
  * Set fertile Cells from placed saved in Params.
  */
  private void setFertileCells() {
      int i;
      int[] coords;
      Cell c;
      
      for (i = 0; i < Params.fertilePlaces.size(); i++) {
          coords = (int[])(Params.fertilePlaces.elementAt(i));
	  c = getCell(coords[0], coords[1]);
	  fertileCells.addElement(c);
	  c.setFertile(true);
      }
  } // setFertileCells()
  
  private void reinitFertileCells() {
      if (fertileCells != null) {
        int c;
        for (c = 0; c < fertileCells.size(); c++)
	    getFertileCell(c).setFertile(false);
        fertileCells.removeAllElements();
      }
  } // reinitFertileCells()
      
  private void createThings(boolean again) {
    boolean userPlaceRocks =
      (Params.getWorld(Params.ROCK_PLACEMENT) == Params.USER_SPECIFIED)
      && Params.rockPlaces.size() == 0;
    
    userPlacePlants =
      (Params.getWorld(Params.PLANT_PLACEMENT) == Params.USER_CONSTRAINED)
      && Params.fertilePlaces.size() == 0;

    if (userPlaceRocks || userPlacePlants) {
	frame.setVisible(true);
        new PickCellFrame(this, ui, userPlaceRocks, userPlacePlants, again);
      } else {
       createRocks();
       setFertileCells();
       createPlants(getNFertileCells() > 0, true);
       createCritters(again);
       frame.setVisible(true);
      }
  } // createThings()
  
  /**
  * Initializes the cells Vector, fills the World with Cells, and
  * sets the neighbors for the Cells.
  */
  private void doCells() {
      Cell cell0, cell1;
      cells = new Cell[Params.getWorld(Params.NCOLUMNS)]
	  [Params.getWorld(Params.NROWS)];
      fill();
      setCellNeighbors();
      cell0 = getCell(0, 0);
      cell1 = getCell((int)(Params.getWorld(Params.NCOLUMNS) - 1),
		      (int)(Params.getWorld(Params.NROWS) - 1));
      //      cell1 = getCell((int)((Params.getWorld(Params.NCOLUMNS) - 1)
      //			    * Sonar.MAX_DISTANCE_PROP), 0);
      //      cell2 = getCell(0, (int)(Params.getWorld(Params.NROWS) - 1));
      Params.maxCellDistance =
	   Pixel.getDistance(cell0.getCenterX(), cell0.getCenterY(),
			     cell1.getCenterX(), cell1.getCenterY());

      Params.minCellDistance = 2 * cellIncr;
  } // doCells()
  
  /**
  * Fills the World with Cells, creating each Cell.
  */
  private void fill() {
    int r, c, x = 0, y = 0;
    Cell cell;

    for (r = 0; r < Params.getWorld(Params.NROWS); r++) {
    	for (c = 0; c < Params.getWorld(Params.NCOLUMNS); c++) {
    	  cells[c][r] = new Cell(ui, c, r, x, y, UI.cellSize);
      	  x += 3 * UI.cellSize;
    	  if (c % 2 == 0)
      	    y += cellIncr;
      	  else
      	    y -= cellIncr;
    	}
    	x = 0;
    	y = (r + 1) * 2 * cellIncr;
    }
  } // fill()
  
  // ACCESS METHODS
  
  /**
  * Returns the Cell in column and row.
  */
  public Cell getCell(int column, int row) {
      if (column >= Params.getWorld(Params.NCOLUMNS))
	  column -= Params.getWorld(Params.NCOLUMNS);
      else if (column < 0)
	  column += Params.getWorld(Params.NCOLUMNS);

      if (row >= Params.getWorld(Params.NROWS))
	  row -= Params.getWorld(Params.NROWS);
      else if (row < 0)
	  row += Params.getWorld(Params.NROWS);

      return (cells[column][row]);
  } // getCell(int, int)
  
  public static int getCellRadius() {
      return 2 * cellIncr;
  } // getCellRadius
  
  /**
  * Returns the Thing with the given index.
  */
  public Thing getThing(int index) {
    return ((Thing) (things.elementAt (index)));
  } // getThing
  
  /**
  * Returns the number of Things in the World.
  */
  public int getNThings() {
      return (things.size());
  } // getNThings

  /**
  * Returns the number of Critters in the World.
  */
  public int getNCritters() {
      return (critters.size());
  } // getNCritters
  
  /**
  * Returns the number of Plants in the World.
  */
  public int getNPlants() {
      return (plants.size());
  } // getNPlants
  
  /**
  * A random row or column.
  */
  private int getRandomRow() {
      return (int)(Math.random() * (Params.getWorld(Params.NROWS)));
  } // getRandomRow()
  
  private int getRandomColumn() {
      return (int)(Math.random() * (Params.getWorld(Params.NCOLUMNS)));
  } // getRandomColumn()
  
  /**
  * Returns 0 (random), 1 (must touch a rock), or 2 (must not touch a rock).
  */
  private int getPlantPlacement() {
      return Params.getWorld(Params.PLANT_PLACEMENT);
  } // getPlantPlacement()
  
  private boolean cellWorks(Cell cell, int placement) {
      if (!cell.isEmpty())
	  return false;
      else if (placement == Params.RANDOM)
	  return true;
      else {
          Cell[] neighbors = cell.getNeighbors();
	  Cell c;
	  Thing t;
	  int i = 0;
	  boolean foundRock = false;

	  while (i < 6 && !foundRock) {
             c = neighbors[i];
             if (c != null) {
               t = c.getContents();
               if (t instanceof Rock)
		  foundRock = true;
             }
            i++;
	  }
	  if (placement == Params.NEXTTO_ROCK)
            return foundRock;
	  else
	    return !foundRock;
      }
  } // cellWorks(Cell, int)
  
  /**
  * A randomly selected unoccupied Cell.
  */
  public Cell getRandomEmptyCell(int depth, int placement) {
      if (depth > SEARCH_LIMIT)
	  return null;
      else {
	  int
	  ranRow = getRandomRow(),
	  ranColumn = getRandomColumn();
	  Cell
	  cell = getCell(ranColumn, ranRow);
	  
	  if (cellWorks(cell, placement))
	     return cell;
	  else
	     return (getRandomEmptyCell(depth + 1, placement));
      }
  } // getRandomEmptyCell(int, int)
  
  /**
  * Returns the Critter with the given index.
  */
  public Critter getCritter(int index) {
      return ((Critter)(critters.elementAt(index)));
  } // getCritter

  /**
  * Returns the Plant with the given index.
  */
  public Plant getPlant(int index) {
      return ((Plant)(plants.elementAt(index)));
  } // getPlant

  /**
  * Just returns the WorldFrame for the World.
  */
  public WorldFrame getFrame() {
      return frame;
  } // getFrame
    
  /**
  * Sets the Cell neighbors in the 6 cardinal directions for each
  * Cell in the World.
  */
  private void setCellNeighbors() {
      int r, c;
      Cell cell;
      
      for (r = 0; r < Params.getWorld(Params.NROWS); r++) {
	  for (c = 0; c < Params.getWorld(Params.NCOLUMNS); c++) {
	      cell = getCell (c, r);
	      setCellNeighbor (cell, Cell.N, c, r-1);
	      setCellNeighbor (cell, Cell.S, c, r+1);
	      if (c % 2 == 0) {
		  setCellNeighbor (cell, Cell.NE, c+1, r-1);
		  setCellNeighbor (cell, Cell.NW, c-1, r-1);
		  setCellNeighbor (cell, Cell.SE, c+1, r);
		  setCellNeighbor (cell, Cell.SW, c-1, r);
	      } else {
		  setCellNeighbor (cell, Cell.NE, c+1, r);
		  setCellNeighbor (cell, Cell.NW, c-1, r);
		  setCellNeighbor (cell, Cell.SE, c+1, r+1);
		  setCellNeighbor (cell, Cell.SW, c-1, r+1);
	      }
	  }
      }
  } // setCellNeighbors

  /**
  * Sets Cell neighbor for cell in direction to be the Cell in the
  * given col and row.
  */
  private void setCellNeighbor (Cell cell, int direction, int col, int row) {
    if (col < 0)
	col = Params.getWorld(Params.NCOLUMNS) - 1;
    else if (col >= Params.getWorld(Params.NCOLUMNS))
	col = 0;
	
    if (row < 0)
	row = Params.getWorld(Params.NROWS) - 1;
    else if (row >= Params.getWorld(Params.NROWS))
	row = 0;
	
//    if (col < 0 || row < 0 || col >= Params.getWorld(Params.NCOLUMNS) ||
//	row >= Params.getWorld(Params.NROWS))
//      cell.setNeighbor (direction, null);
//    else
      cell.setNeighbor(direction, getCell(col, row));
  } // setCellNeighbor
  
  /**
  * This gets called in Critter (and subclasses) when a Critter needs to
  * be animated for something it did or that happened to it.   
  */
  public void setNAnimationSteps (int nSteps) {
      nAnimationSteps = nSteps;
  } // setNAnimationSteps(int)
  
  /**
  * Creates a Rock forming a square if proportion is
  * greater than 0.   
  */
  public void makeBarrier(int proportion) {
      if (proportion > 0) {
        Rock barrier = new Rock(this, ui, Rock.COLOR);
	int lastC = ((Params.getWorld(Params.NCOLUMNS) - 1) * proportion) / 10;
	int lastR = ((Params.getWorld(Params.NROWS) - 1) * proportion) / 10;
        int r, c;
	
	// Left and right sides
	for (r = 0; r <= lastR; r++) {
	    barrier.addCell(getCell(0, r));
	    barrier.addCell(getCell(lastC, r));
	}
	// Top and bottom
	for (c = 1; c < lastC; c++) {
	    barrier.addCell(getCell(c, 0));
	    barrier.addCell(getCell(c, lastR));
	}
      }
  } // makeBarrier(int)

  /**
  * Adds thing to the things Vector and returns an index.
  */
  public int addThing (Thing thing) {
    things.addElement (thing);
    if (thing instanceof Plant)
      nPlants++;
    return (things.size() - 1);
  } // addThing(Thing)
  
  /**
  * Just adds the Critter to the list (Vector) of critters
  * and returns the critter index (passed to critter).
  */
  public int addCritter (Critter critter) {
    critters.addElement (critter);
    ui.write("Adding critter " + (critters.size() - 1), 1);
    return (critters.size() - 1);
  } // addCritter
  
  /**
  * Removes the things from the World, making necessary
  * adjustments if it's a Critter or Plant.
  */
  public void removeThing(Thing thing) {
    things.removeElement (thing);
    if (thing instanceof Critter)
	((Critter)thing).makeGone();
//      critters.removeElement (thing);
    else if (thing instanceof Plant) {
      plants.removeElement (thing);
      nPlants--;
    }
  } // removeThing
  
  private Cell getFertileCell(int i) {
      return (Cell)(fertileCells.elementAt(i));
  } // getFertileCell(int)
  
  private int getNFertileCells() {
      return fertileCells.size();
  } // getNFertileCells()

  private Cell getEmptyFertileCell() {
      int c = 0;
      Cell cell = null;
      boolean empty = false;
      while (!empty && c < getNFertileCells()) {
          if (getFertileCell(c).isEmpty())
	      empty = true;
	  else
	      c++;
      }
      if (empty) {
        c = (int)(Math.random() * getNFertileCells());
	empty = false;
      
        while (!empty) {
          cell = getFertileCell(c);
          if (cell.isEmpty()) {
	      empty = true;
          }
	  else {
	      c = (int)(Math.random() * getNFertileCells());
	  }
        }
        return cell;
      } else
	  return null;
  } // getEmptyFertileCell()
  
  private Plant createPlant(Cell cell) {
      Plant plant = new Plant(this, ui);
      plant.addCell(cell);
      
      return plant;
  } // createPlant(Cell)

  /**
  * Adds a new Plant to a randomly selected empty Cell.
  */
  private Plant createPlant(boolean inFertile, boolean first) {
      Cell cell;
      Plant plant = null;
      
      if (inFertile)
	  cell = getEmptyFertileCell();
      else
	  cell = getRandomEmptyCell(0, getPlantPlacement());
      
      if (cell != null) {
	  plant = new Plant(this, ui);
	  plant.addCell(cell);
	  if (!first)
  	    plant.paint();
      }
      return plant;
  } // createPlant(boolean, boolean)
  
  /**
  * Just adds the plant to the plants vector.
  */
  public void addPlant (Plant plant) {
      plants.addElement(plant);
  } // addPlant
  
  /**
  * Creates enough new Plants to reach Params.initNPlants.
  */
  public void createPlants(boolean inFertile, boolean first) {
      boolean isRoom = true;
      while (nPlants < Params.getWorld(Params.NPLANTS) && isRoom) {
          if (createPlant(inFertile, first) == null)
	      isRoom = false;
      }
  } // createPlants(boolean, boolean)
  
  public void createCritters(boolean again) {
      int speciesI = 0, critI;
      Cell cell;
      boolean isRoom = true;
      
      for (speciesI = 0;
	   speciesI < Params.getNSpecies() && isRoom;
	   speciesI++) {
	  for (critI = 0;
	       critI < Params.getSpecies(speciesI).get(Params.N) && isRoom;
	       critI++) {
	      cell = getRandomEmptyCell(0, 0);
	      if (cell != null)
		  new Critter(this, ui, speciesI).setCell(cell);
	      else {
		  isRoom = false;
	      }
	  }
     }

     initFrames(again, Params.getNSpecies() > 0);
//     frame.repaint();
  } // createCritters()
  
  public void createSpecies(int speciesI) {
      int critI;
      boolean isRoom = true;
      Cell cell;
      Critter critter;
      Species spec = Params.getSpecies(speciesI);
      
      for (critI = 0;
	   critI < spec.get(Params.N) && isRoom;
	   critI++) {
	  cell = getRandomEmptyCell(0, 0);
	  if (cell != null) {
	      critter = new Critter(this, ui, speciesI);
	      critter.setCell(cell);
	  } else {
	      isRoom = false;
	  }
      }
      
  } // createSpecies(int)
  
  /**
  * Creates a Rock in a particular Cell.
  */
  private Rock createRock(Cell cell) {
      Rock rock = new Rock(this, ui);
      rock.addCell(cell);
      
      return rock;
  } // createRock(Cell)
  
  /**
  * Creates a Rock of ideally size Cells, possibly fewer,
  * in a randomly selected position.
  */
  private Rock createRock(int size) {
      Rock rock = null;
      Cell cell = getRandomEmptyCell(0, 0);
      int nCells = 0;
      
      if (cell != null) {
	  rock = new Rock(this, ui);
	  while (nCells < size && cell != null) {
	      rock.addCell (cell);
	      cell = cell.getRandomEmptyNeighbor (rock);
	      nCells++;
	  }
      }
      return rock;
  } // createRock(int)
  
  /**
  * Creates Params.getWorld(Params.NROCKS) Rocks or lets the user pick
  * where the rocks should go.   
  */
  public void createRocks() {
      int rockI = 0;
      
      if (Params.rockPlaces.size() > 0) {
          // Reinstate saved rocks.
	  int[] coords;
	  for (rockI = 0; rockI < Params.rockPlaces.size(); rockI++) {
	      coords = (int[])(Params.rockPlaces.elementAt(rockI));
	      new Rock(this, ui).addCell(getCell(coords[0], coords[1]));
	  }
	  
      } else {
	  boolean isRoom = true;
	  
	  // First make the barrier.
          makeBarrier(Params.getWorld(Params.BARRIER));

	  // Then the randomly created rocks.
	  while (rockI < Params.getWorld(Params.NROCKS) && isRoom) {
	      if (createRock(1) == null)
		  isRoom = false;
	      rockI++;
	  }
      }
  } // createRocks()

  /**
  * Assigns the canvas to the Cells in the World.
  * This can't be done when the Cells are created because the canvas
  * doesn't exist yet.
  */
  public void addCanvas (WorldCanvas canvas) {
    for (int c = 0; c < Params.getWorld(Params.NCOLUMNS); c++)
      for (int r = 0; r < Params.getWorld(Params.NROWS); r++)
         getCell (c, r).canvas = canvas;
  } // addCanvas(WorldCanvas)
    
  /**
  * paints the cleared World.
  */
  public void clear () {
      canvas.clear();
  } // clear
  
  /**
  * Several methods for painting the World.  This involves painting
  * the Things in the World and the paths behind Critters.
  */
  public void paint(Graphics graphics) {
      canvas.clear(graphics);
      paintCells(graphics);
      paintThings(graphics);
  } // paint
  
  public void paint() {
      paint(canvas.getGraphics());
  } // paint
  
  public void paintThings(Graphics graphics) {
    for (int i = 0; i < things.size(); i++) {
//	ui.write("Painting thing " + i);
      getThing(i).paint(graphics);
    }
  }  // paintThings
  
  public void paintCells(Graphics graphics) {
    int r, c;
    for (r = 0; r < Params.getWorld(Params.NROWS); r++) {
        for (c = 0; c < Params.getWorld(Params.NCOLUMNS); c++) {
	  getCell(c, r).paint(graphics);
      }
    }
  } // paintCells(Graphics)
  
  /**
  * (Un)highlight the Critter with the given index.
  */
  public void showCritter(int critterIndex, boolean show) {
      Critter critter = getCritter (critterIndex);
      if (!critter.isGone()) {
        Cell cell = critter.getCell();
      
        if (show)
          cell.circleFill(HIGHLIGHT_COLOR);
        else
          critter.paint(cell);
      }
  } // showCritter
  
  /**
  * Writes the state of the World in the UI.
  */
  public void describe() {
    ui.write (name + ":");
    ui.write ("  " + Params.getWorld(Params.NCOLUMNS) + " columns");
    ui.write ("  " + Params.getWorld(Params.NROWS) + " rows");
    for (int i = 0; i < things.size(); i++)
      getThing (i).describe();
  } // describe
  
  /**
  * Runs the World.  All of the work is done in run().
  */
  public void start() {
      thread = new Thread(this);
      thread.start();
  } // start()
  
  public void interrupt() {
      if (thread != null) {
	  thread.stop();
	  showRun();
      }
  } // interrupt()
  
  private void initRun() {
     for (int c = 0; c < critters.size(); c++)
        getCritter(c).initRun();
  } // initRun()
  
  /**
  * This is the run() which gets called when thread.start() is called.
  * Runs the World for Params.runTimeSteps time steps.  Brings the WorldFrame to
  * the front, then repeatedly runs step until either nothing has changed
  * or Params.runTimeSteps has elapsed.
  */
  public void run() {
      int tStep, c;
      boolean alive = true;
        
      frame.toFront();
      thread.yield();
      
      initRun();
            
      // Step each of the Critters for Params.runTimeSteps time steps.
      for (tStep = 0;
	   tStep < UI.nRunTimeSteps && alive;
	   tStep++) {
	  ui.write ("Step " + tStep, 1);
	  alive = step(tStep);
	  pause(UI.stepPause);
      }
      
      showRun();
      
      //      thread.stop();
  } // run()
  
  private void showRun() {
     PrintWriter trans = null;
     
     if (ui.makeTranscript())
	 trans = ui.openTranscript();
     
     repaintCritterFrame(STRENGTH);
     repaintCritterFrame(REINFORCEMENT);
     repaintCritterFrame(LTM);

     ui.write("The World is " + totalTimeSteps + " time steps old.");

     for (int s = 0; s < Params.getNSpecies(); s++) {
	 if (trans != null) 
	     trans.println("Results after " + totalTimeSteps);
	 Params.getSpecies(s).describeMembers(trans, totalTimeSteps);
	 if (trans != null)
	     trans.println();
     }
     
     if (trans != null)
	 trans.close();

  } // showRun()

  /**
  * Steps the world by stepping each Critter.  If tstep is negative, we're
  * in step mode and animation and showing rays is possible.   
  */
  public boolean step(int tStep) {
      int c;
      boolean alive = false;
      
      if (tStep < 0)
	  initRun();

      // Recreate Plants if Plants are too few.
      if (nPlants == 0  || (maintainNPlants && nPlants < Params.getWorld(Params.NPLANTS))) {
        createPlants(getNFertileCells() > 0, false);
      }

      // Step each Critter.
      for (c = 0; c < critters.size(); c++) {
	  if (getCritter(c).step())
	      alive = true;
      }

      showStep(tStep);
      
      if (tStep < 0) {
	  if (ui.animate())
            animate();
          if (ui.animationEnabled) {
      	    pause(ui.withinStepPause);
	    paint();
          }
      }
      
      if (alive)
        totalTimeSteps++;
      
      return alive;
  } // step
  
  private void showStep(int tStep) {
      
      repaintCritterFrame(STM);
      
//    if (memoryFrame != null)
//      memoryFrame.repaint();

  if (tStep < 0 || (tStep > 0 && tStep % SHOW_STATS_FREQ == 0)) {
      repaintCritterFrame(STRENGTH);
      repaintCritterFrame(REINFORCEMENT);
    }
    
  if (tStep < 0)
      repaintCritterFrame(LTM);
  } // showStep(int)
  
  /**
  * Animate all Critters that say they should be (their animateType >= 0).
  */
  private void animate() {
      // Only do animation if some Critter wants it.
      if (nAnimationSteps > 0) {
        int c, step;
	for (step = nAnimationSteps; step > 0; step--) {
	    for (c = 0; c < critters.size(); c++)
		getCritter (c).animate(step);
	    if (step > 1)
		pause (Params.animationPause);
	}
        // Reset the number of animation steps for the next run time step.
        nAnimationSteps = 0;
      }
  } // animate()
  
  /**
  * Pauses for 10 * length milliseconds.
  */
  private void pause(int length) {
    try {thread.sleep (length * 10);}
    catch (InterruptedException e) {
      System.err.println ("pause: " + e); }
  } // pause(int)
  
  private void writeParam(PrintWriter out, int i) {
      out.println(Params.getWorldKey(i) + " " + Params.getWorld(i));
  }
  
  private void writeParam(PrintWriter out, int i, int value) {
      out.println(Params.getWorldKey(i) + " " + value);
  }
  
  public void write(String filename, String directory) {
    PrintWriter out = null;
    String pathname = directory + filename;
//    String pathname = UI.makePathname(filename, Params.DIRECTORY);
    Thing t;
    Cell c;
    int i;

    try {
      out = new PrintWriter(new FileOutputStream(pathname));
      ui.write ("Saving environment in " + filename + ".");

      /*
      if (Params.envFilename != null)
	  out.println(Params.ENVIRONMENT_KEY + " " + Params.envFilename);
      */
      
      for (i = 0; i < Params.NWORLD_PARAMS; i++) {
	  if (i == Params.NROCKS || i == Params.BARRIER)
	      // Never write these: rocks are specified explicitly and barriers
	      // aren't needed anymore.
	      ;
	  else if (i == Params.EATEN_DECR || i == Params.CRITTER_EATING_INCR) {
	      if (Params.getWorld(Params.CRITTER_EDIBILITY) == Params.EDIBLE)
		  writeParam(out, i);
	  }
	  else
    	      writeParam(out, i);
      }

      for (i = 0; i < getNThings(); i++) {
	  t = getThing(i);
	  if (t instanceof Rock)
	      ((Rock)t).write(out);
      }

      if (fertileCells != null)
	  for (i = 0; i < getNFertileCells(); i++) {
	      c = getFertileCell(i);
	      out.println("fertile  " + c.getColumn() + " " + c.getRow());
	  }
	      
      out.close();
      
    } catch (IOException e) {
	System.err.println ("World: write " + e);
    }
  }

  public void write(String filename, String directory, Critter c, String name) {
      PrintWriter out = null;
      String pathname = directory + filename;
      int i;
      
      try {
	  out = new PrintWriter(new FileOutputStream(pathname));
	  ui.write ("Saving critter in " + filename + ".");
	  
	  for (i = 0; i < Params.NCRITTER_PARAMS; i++) {
	      if (i != Params.N && i != Params.NAME &&
		  i != Params.NCOLORS && i != Params.NSONAR_DISTINCTIONS)
  	        c.writeParam(out, i);
	  }
	 
	  // Write out the name, age, and memory
	  out.println("name  " + name);
	  c.writeAge(out);
	  c.writeMemory(out);
	  
	  out.close();
	  
      } catch (IOException e) {
	  System.err.println ("World: write " + e);
      }
  } // write(String, Critter)
  
} // class World
