/** 
 * SALSA 2
 * July 2000
 * Mike Gasser (gasser@indiana.edu)
 **/

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/**
* The Frame in which the World gets drawn.  The drawing actually takes
* place in the WorldCanvas that gets created when WorldFrame is
* instantiated.  This is a MouseListener for the case
* when CreateThingFrame has been instantiated and Things are being
* created by clicking in the Frame or when a Critter is being
* selected after the "Highlight Critter" button is pushed.
*/

public class WorldFrame extends Frame
  implements WindowListener, MouseListener {

  public final static Color
    BG_COLOR = Color.black;
    
  private UI ui;
  private World world;

  private WorldCanvas canvas;
  
  private Vector[] selectedCells = null;
  
  private Vector
    removeThings = null,
    removeFertileCells = null;
  
  // Variables required for selection of Cells when new Things are created
  // using CreateThingFrame.
  private Vector
    selectedRockCells,
    selectedPlantCells;

  private Color selectedColor;
  
  private Critter selectedCritter;
  
  private int selectedType;
  
  private boolean
    anySelected = false,
    selectCritter = false,
    selectPlantCells = false,
    selectCells = false;

  private String helpText =
    "This window displays the world and the things in it.";
    
  WorldFrame (World world, UI ui, String name, int columns, int rows,
              int cellSize, int cellIncr) {
    this.world = world;
    this.ui = ui;

    canvas = new WorldCanvas (world, ui, columns, rows, cellSize, cellIncr);
    canvas.addMouseListener (this);
    setLayout (new GridLayout (1, 1));
    add (canvas);
    
    selectedRockCells = new Vector();
    selectedPlantCells = new Vector();

    addWindowListener (this);
    addMouseListener (this);

    setBackground (BG_COLOR);
    setTitle (name);
    pack();
    setVisible (false);
  } // constructor WorldFrame
  
  private void removeCell(Cell cell, boolean plant) {
      cell.clear();
      if (plant)
        selectedPlantCells.removeElement(cell);
      else
	selectedRockCells.removeElement(cell);
  } // removeCell(Cell, boolean)
  
  private void addCell(Cell cell, boolean plant) {
      cell.smallCircleFill(selectedColor);
      if (plant)
        selectedPlantCells.addElement(cell);
      else
	selectedRockCells.addElement(cell);
      anySelected = true;
  } // addCell(Cell, boolean)
  
  private void removeCell(Cell cell) {
      Vector v = selectedCells[selectedType];

      cell.clear();
      v.removeElement(cell);
  } // removeCell(Cell, int)
  
  private void removeCellFromOther(Cell cell) {
      Vector v;
      
      cell.clear();
      for (int i = 0; i < selectedCells.length; i++)
	  if (i != selectedType) {
	      v = selectedCells[i];
	      v.removeElement(cell);
	  }
  } // removeCellFromOther(Cell)
  
  private void addCell(Cell cell) {
      cell.smallCircleFill(selectedColor);
      selectedCells[selectedType].addElement(cell);
      anySelected = true;
  } // addCell(Cell, int)
  
  private void addRemoveThing(Thing thing, Cell cell) {
      cell.hide();
      thing.outline(cell);
      removeThings.addElement(thing);
  } // addRemoveThing(Thing, Cell)
  
  private void replaceThing(Thing thing, Cell cell) {
      thing.paint(cell);
      removeThings.removeElement(thing);
  } // replaceThing(Thing, Cell)
  
  private boolean inSelected(Cell cell) {
      return selectedCells[selectedType].contains(cell);
  } // inSelected(Cell)
  
  private boolean inOtherSelected(Cell cell) {
      boolean in = false;
      
      for (int i = 0; i < selectedCells.length && !in; i++)
	  if (i != selectedType && selectedCells[i].contains(cell))
	      in = true;

      return in;
  } // inOtherSelected(Cell)
  
  public boolean anySelectedCells() {
      return anySelected;
  } // anySelectedCells()

  public void mouseClicked(MouseEvent event) {
    if (selectCells || selectCritter) {
        
      int coords[] = {event.getX(), event.getY()};
      Cell selectedCell = canvas.getSelectedCell(coords);
      
      if (selectCritter) {
          if (selectedCell.isEmpty())
	      // Do nothing if there's nothing there
	      ;
	  else {
	      Thing t = selectedCell.getContents();
	      if (t != null && t instanceof Critter) {
		  selectedCritter = (Critter)t;
	      }
	  }
	  
      } else {
	  // selectCells == t

	  if (selectedCell.isEmpty()) {
	      if (inSelected(selectedCell))
		  removeCell(selectedCell);
		  
	      // Cell is already on other list
	      else if (inOtherSelected(selectedCell)) {
		  removeCellFromOther(selectedCell);
		  addCell(selectedCell);
	      } else
		  addCell(selectedCell);

	  } else {
	      // selectedCell is not empty
	      Thing t = selectedCell.getContents();
	      if (removeThings.contains(t)) {
//		  ui.write("Replacing thing.");
		  replaceThing(t, selectedCell);
	      } else {
//		  ui.write("Adding thing to remove list.");
		  addRemoveThing(t, selectedCell);
	      }
	  }
      }
    }
  } // method mouseClicked

  public void mousePressed (MouseEvent event) {
  } // method mousePressed
    
  public void mouseReleased (MouseEvent event) {
  } // method mouseReleased

  public void mouseEntered (MouseEvent event) {
    if (ui.showHelp) {
      ui.helpWrite (helpText);
    }
  } // method mouseEntered
  
  public void mouseExited (MouseEvent event) {
    if (ui.showHelp) {
      ui.helpWrite (UI.HELP_INSTRUCTIONS);
    }
  } // method mouseExited

  public void windowClosed (WindowEvent event) {
  } // method windowClosed

  public void windowDeiconified (WindowEvent event) {
  } // method windowDeiconified

  public void windowIconified (WindowEvent event) {
  } // method windowIconified

  public void windowActivated (WindowEvent event) {
  } // method windowActivated

  public void windowDeactivated (WindowEvent event) {
  } // method windowDeactivated

  public void windowOpened (WindowEvent event) {
  } // method windowOpened

  public void windowClosing (WindowEvent event) {
    ui.removeWindow (this);
  } // method windowClosing
  
  /**
  * Returns the index'th Cell in selectedCells.
  */
  public Cell getSelectedCell(int index, boolean plant) {
      if (plant)
	  return ((Cell)(selectedPlantCells.elementAt(index)));
      else
        return ((Cell)(selectedRockCells.elementAt (index)));
  } // method getSelectedCell(int)
  
  public Cell getSelectedCell(int index, int type) {
      return ((Cell)(selectedCells[type].elementAt(index)));
  } // getSelectedCell(int, int)
  
  public Vector[] getSelectedCells() {
      return selectedCells;
  }
  
  public Vector getRemoveThings() {
      return removeThings;
  }
  
  public Vector getRemoveFertileCells() {
      return removeFertileCells;
  }
  
  public Vector getSelectedCells(boolean plant) {
      if (plant)
	  return selectedPlantCells;
      else
	  return selectedRockCells;
  } // getSelectedCells(boolean)
  
  public int getNSelectedCells(boolean plant) {
      if (plant)
	  return selectedPlantCells.size();
      else
          return selectedRockCells.size();
  } // getNSelectedCells()
  
  public Critter getSelectedCritter() {
      return selectedCritter;
  } // getSelectedCritter()
  
  public boolean critterSelected() {
      return (selectedCritter != null) ;
  } // critterSelected()
  
  public boolean critterTypeSelected() {
      return selectedType >= EditWorldFrame.CRITTER;
  } // critterTypeSelected()
  
  /**
  * Selects a color.
  */
  public void select(Color color, int type) {
      selectedColor = color;
      selectedType = type;
      selectPlantCells =
        (type != EditWorldFrame.ROCK && type != EditWorldFrame.CRITTER);
  } // select(Color, int)
  
  /**
  * Gets the Frame ready for selection of Cells in Things to be
  * created.  Called in EditWorldFrame.
  */
  public void initSelect() {
    initSelectedCellsAndThings();
    selectedRockCells.removeAllElements();
    selectedPlantCells.removeAllElements();
    selectCells = true;
    setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
    anySelected = false;
  } // method initSelect
  
  private void initSelectedCellsAndThings() {
      int v;
      
      if (selectedCells == null) {
	  selectedCells = new Vector[EditWorldFrame.nTypes];
	  for (v = 0; v < EditWorldFrame.nTypes; v++)
	      selectedCells[v] = new Vector();
      } else {
	  for (v = 0; v < EditWorldFrame.nTypes; v++)
	      selectedCells[v].removeAllElements();
      }

      if (removeThings == null) {
	  removeThings = new Vector();
      } else {
	  removeThings.removeAllElements();
      }

      if (removeFertileCells == null) {
	  removeFertileCells = new Vector();
      } else {
	  removeFertileCells.removeAllElements();
      }
  } // initSelectedCellsAndThings()
  
  public void initSelectCritter() {
      selectedCritter = null;
      selectCritter = true;
  } // initSelectCritter()

  /**
  * Turns off selection of Cells for creating Things.  Called in
  * CreateThingFrame.
  */
  public void deinitSelect() {
    selectedRockCells.removeAllElements();
    selectedPlantCells.removeAllElements();
    selectCells = false;
    setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  } // method deinitSelect()
  
  public void deinitSelectCritter() {
      selectCritter = false;
  } // deinitSelectCritter()
  
  public void repaint() {
      canvas.repaint();
  } // repaint()
  
} // class WorldFrame
